package com.adobe.ac.ant.tasks;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.text.MessageFormat;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.XMLWriter;

public class FlexUnitTask
      extends Task
{

   private static final String END_OF_TEST_RUN = "<endOfTestRun/>";
   private static final String END_OF_TEST_SUITE = "</testsuite>";
   private static final String END_OF_TEST_ACK = "<endOfTestRunAck/>";
   private static final char NULL_BYTE = '\u0000';
   private static final String FILENAME_PREFIX = "TEST-";
   private static final String FILENAME_EXTENSION = ".xml";

   private static final String POLICY_FILE_REQUEST = "<policy-file-request/>";
   final static String DOMAIN_POLICY = "<cross-domain-policy><allow-access-from domain=\"*\" to-ports=\"{0}\" /></cross-domain-policy>";

   private boolean failures = false;
   private boolean complete;
   private String lastFailureMessage;
   private boolean verbose = true;

   // attributes from ant task def
   private int port = 1024;
   private int socketTimeout = 60000; // milliseconds
   private boolean failOnTestFailure = true;
   private String failureProperty = "flexunit.failed";
   private String swf;
   private File reportDir;

   /**
    * Set the port to receive the test results on.
    * 
    * @param serverPort the port to set.
    */
   public void setPort(
         final int serverPort )
   {
      port = serverPort;
   }

   /**
    * Set the timeout for receiving the flexunit report.
    * 
    * @param timeout in milliseconds.
    */
   public void setTimeout(
         final int timeout )
   {
      socketTimeout = timeout;
   }

   /**
    * The SWF for the FlexUnit tests to run.
    * 
    * @param testSWF the SWF to set.
    */
   public void setSWF(
         final String testSWF )
   {
      swf = testSWF;
   }

   /**
    * Set the directory to output the test reports to.
    * 
    * @param toDir the directory to set.
    */
   public void setToDir(
         final String toDir )
   {
      reportDir = new File( toDir );
   }

   /**
    * Should we fail the build if the flex tests fail?
    * 
    * @param fail
    */
   public void setHaltonfailure(
         final boolean fail )
   {
      failOnTestFailure = fail;
   }

   /**
    * Custom ant property noting test failure
    * 
    * @param failprop
    */
   public void setFailureproperty(
         final String failprop )
   {
      failureProperty = failprop;
   }

   /**
    * Custom ant property noting test failure
    * 
    * @param failprop
    */
   public void setVerbose(
         final boolean v )
   {
      verbose = v;
   }

   /**
    * Called by Ant to execute the task.
    */
   public void execute() throws BuildException
   {
      // Check a SWF was specified.
      if ( swf == null
            || swf.length() == 0 )
      {
         throw new BuildException( "You must specify the 'swf' property." );
      }

      // Check we have an output dir.
      if ( reportDir == null )
      {
         reportDir = getProject().resolveFile(
               "." );
      }
      reportDir.mkdirs();

      // Start a thread that receives the FlexUnit results.
      receiveFlexUnitResults();

      // Start the browser and run the FlexUnit tests.
      final FlexUnitLauncher browser = new FlexUnitLauncher();
      try
      {
         browser.runTests( swf );
      }
      catch ( Exception e )
      {
         throw new BuildException( "Error launching the test runner.", e );
      }

      // Wait until the tests are complete.
      while ( !complete )
      {
         try
         {
            Thread.sleep( 1000 );
         }
         catch ( InterruptedException e )
         {
            throw new BuildException( e );
         }
      }

      handleFailures();
   }

   /**
    * Create a server socket for receiving the test reports from FlexUnit. We
    * read the test reports inside of a Thread.
    */
   private void receiveFlexUnitResults() throws BuildException
   {
      // Start a thread to accept a client connection.
      final Thread thread = new Thread()
      {
         private ServerSocket serverSocket = null;
         private Socket clientSocket = null;
         private InputStream in = null;
         private OutputStream out = null;

         public void run() throws BuildException
         {
            try
            {
               openServerSocket();
               openClientSocket();

               StringBuffer buffer = new StringBuffer();
               int bite = -1;

               while ( ( bite = in.read() ) != -1 )
               {
                  final char chr = ( char ) bite;

                  if ( chr == NULL_BYTE )
                  {
                     final String data = buffer.toString();
                     buffer = new StringBuffer();

                     if ( data.equals( POLICY_FILE_REQUEST ) )
                     {
                        sendPolicyFile();
                        closeClientSocket();
                        openClientSocket();
                     }
                     else if ( data.endsWith( END_OF_TEST_SUITE ) )
                     {
                        saveTestReport(data);
                     }
                     else if ( data.equals( END_OF_TEST_RUN ) )
                     {
                        sendAcknowledgement();
                     }
                  }
                  else
                  {
                     buffer.append( chr );
                  }
               }
            }
            catch ( BuildException be )
            {
               try
               {
                  sendAcknowledgement();
               }
               catch ( IOException e )
               {
                  // ignore
               }
               
               lastFailureMessage = be.getMessage();
            }
            catch ( SocketTimeoutException e )
            {
               throw new BuildException( "timeout waiting for flexunit report",
                     e );
            }
            catch ( IOException e )
            {
               throw new BuildException(
                     "error receiving report from flexunit", e );
            }
            finally
            {
               // always stop the server loop
               complete = true;

               closeClientSocket();
               closeServerSocket();
            }
         }

         private void sendPolicyFile() throws IOException
         {
            out.write( MessageFormat.format(
                  DOMAIN_POLICY, new Object[]
                  { Integer.toString( port ) } ).getBytes() );

            out.write( NULL_BYTE );

            if ( verbose )
            {
               log( "sent policy file" );
            }
         }

         private void saveTestReport(
               final String report )
         {
            writeTestReport( report );

            if ( verbose )
            {
               log( "end of test" );
            }
         }

         private void sendAcknowledgement() throws IOException
         {
            out.write( END_OF_TEST_ACK.getBytes() );
            out.write( NULL_BYTE );

            if ( verbose )
            {
               log( "end of test run" );
            }
         }

         private void openServerSocket() throws IOException
         {
            serverSocket = new ServerSocket( port );
            serverSocket.setSoTimeout( socketTimeout );

            if ( verbose )
            {
               log( "opened server socket" );
            }
         }

         private void closeServerSocket()
         {
            if ( serverSocket != null )
            {
               try
               {
                  serverSocket.close();
               }
               catch ( IOException e )
               {
                  // ignore
               }
            }
         }

         private void openClientSocket() throws IOException
         {
            // This method blocks until a connection is made.
            clientSocket = serverSocket.accept();

            if ( verbose )
            {
               log( "accepting data from client" );
            }

            in = clientSocket.getInputStream();
            out = clientSocket.getOutputStream();
         }

         private void closeClientSocket()
         {
            // Close the output stream.
            if ( out != null )
            {
               try
               {
                  out.close();
               }
               catch ( IOException e )
               {
                  // ignore
               }
            }

            // Close the input stream.
            if ( in != null )
            {
               try
               {
                  in.close();
               }
               catch ( IOException e )
               {
                  // ignore
               }
            }

            // Close the client socket.
            if ( clientSocket != null )
            {
               try
               {
                  clientSocket.close();
               }
               catch ( IOException e )
               {
                  // ignore
               }
            }
         }
      };

      thread.start();
   }

   /**
    * Write a test report to disk.
    * 
    * @param report the report to write.
    */
   private void writeTestReport(
         final String report )
   {
      try
      {
         // Parse the report.
         final Document document = DocumentHelper.parseText( report );

         // Get the test attributes.
         final Element root = document.getRootElement();
         final String name = root.valueOf( "@name" );
         final int numFailures = Integer.parseInt( root.valueOf( "@failures" ) );

         if ( verbose )
            log( "Running "
                  + name );
         if ( verbose )
            log( formatLogReport( root ) );

         // Get the output file name.
         final File file = new File( reportDir, FILENAME_PREFIX
               + name + FILENAME_EXTENSION );

         // Pretty print the document to disk.
         final OutputFormat format = OutputFormat.createPrettyPrint();
         final XMLWriter writer = new XMLWriter( new FileOutputStream( file ),
               format );
         writer.write( document );
         writer.close();

         // First write the report, then fail the build if the test failed.
         if ( numFailures > 0 )
         {
            failures = true;

            if ( verbose )
            {
               log( "flexunit test "
                     + name + " failed." );
            }

            if ( failOnTestFailure )
            {
               throw new BuildException( "flexunit test "
                     + name + " failed." );
            }
         }

      }
      catch ( BuildException be )
      {
         throw be;
      }
      catch ( Exception e )
      {
         throw new BuildException( "error writing report to disk", e );
      }
   }

   /**
    * crafts a simple junit type log message.
    */
   private String formatLogReport(
         final Element root )
   {
      int numFailures = Integer.parseInt( root.valueOf( "@failures" ) );
      int numErrors = Integer.parseInt( root.valueOf( "@errors" ) );
      int numTests = Integer.parseInt( root.valueOf( "@tests" ) );
      int time = Integer.parseInt( root.valueOf( "@time" ) );

      final StringBuffer msg = new StringBuffer();
      msg.append( "Tests run: " );
      msg.append( numTests );
      msg.append( ", Failures: " );
      msg.append( numFailures );
      msg.append( ", Errors: " );
      msg.append( numErrors );
      msg.append( ", Time Elapsed: " );
      msg.append( time );
      msg.append( " sec" );

      return msg.toString();
   }

   /**
    * Failure handling. Write ant property for use later in build script.
    */
   private void handleFailures()
   {
      if ( failures )
      {
         getProject().setNewProperty(
               failureProperty, "true" );
         if ( failOnTestFailure )
         {
            throw new BuildException( lastFailureMessage );
         }
      }
   }

   public void log(
         final String message )
   {
      System.out.println( message );
   }

   public static void main(
         String[] args )
   {
      final FlexUnitTask test = new FlexUnitTask();
      test.setTimeout( 0 );
      test
            .setSWF( "C:/dev/swat/projects/ac_emea/FlexUnitExample/bin/AntTestRunner.swf" );
      test.setToDir( "reports" );
      test.execute();
   }
}
